-------------------------------------------------------------------------------
--! @file   CipherCore_Datapath.vhd
--! @brief  Datapath unit for Tiaoxin
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.AES_pkg.all;

entity CipherCore_Datapath is
    generic (
        G_REFERENCE      : boolean := False
    );
    port    (
        clk             : in  std_logic;
        rst             : in  std_logic;

        --! Input
        bdi             : in  std_logic_vector(256              -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(256/8            -1 downto 0);
        key             : in  std_logic_vector(32               -1 downto 0);

        --! Control
        len_a           : in  std_logic_vector(64               -1 downto 0);
        len_d           : in  std_logic_vector(64               -1 downto 0);
        en_key          : in  std_logic;
        en_state        : in  std_logic;
        sel_init        : in  std_logic;
        sel_data        : in  std_logic_vector(3                -1 downto 0);
        sel_decout      : in  std_logic;
        sel_tag         : in  std_logic;

        --! Output
        bdo             : out std_logic_vector(256              -1 downto 0);
        msg_auth_valid  : out std_logic
    );
end entity CipherCore_Datapath;

architecture structure of CipherCore_Datapath is
    signal b0           : std_logic_vector(128                  -1 downto 0);
    signal b1           : std_logic_vector(128                  -1 downto 0);
    signal b01          : std_logic_vector(128                  -1 downto 0);
    signal reg_key      : std_logic_vector(128                  -1 downto 0);
    signal tag          : std_logic_vector(128                  -1 downto 0);
    type t_state is array (integer range <>)
                 of std_logic_vector(128     -1 downto 0);
    signal reg_t3, t3   : t_state(0 to 2);
    signal reg_t4, t4   : t_state(0 to 3);
    signal reg_t6, t6   : t_state(0 to 5);

    constant ZEROS      : std_logic_vector(128                  -1 downto 0)
                        := (others => '0');
    constant Z0         : std_logic_vector(128                  -1 downto 0)
                        := x"428a2f98d728ae227137449123ef65cd";
    constant Z1         : std_logic_vector(128                  -1 downto 0)
                        := x"b5c0fbcfec4d3b2fe9b5dba58189dbbc";
begin
    pRegs:
    process(clk)
    begin
        if rising_edge(clk) then
            if (en_key = '1') then
                reg_key <= reg_key(95 downto 0) & key;
            end if;
            if (en_state = '1') then
                if (sel_init = '1') then
                    reg_t3(0) <= reg_key;
                    reg_t3(1) <= reg_key;
                    reg_t3(2) <= b0;

                    reg_t4(0) <= reg_key;
                    reg_t4(1) <= reg_key;
                    reg_t4(2) <= b0;
                    reg_t4(3) <= Z0;

                    reg_t6(0) <= reg_key;
                    reg_t6(1) <= reg_key;
                    reg_t6(2) <= b0;
                    reg_t6(3) <= Z1;
                    reg_t6(4) <= (others =>'0');
                    reg_t6(5) <= (others =>'0');
                else
                    reg_t3 <= t3;
                    reg_t4 <= t4;
                    reg_t6 <= t6;
                end if;
            end if;
        end if;
    end process;

    b0  <= bdi(255 downto 128);
    b1  <= bdi(127 downto 0);
    b01 <= b0 xor b1;
    
    tag <= reg_t3(0) xor reg_t3(1) xor reg_t3(2) xor
           reg_t4(0) xor reg_t4(1) xor reg_t4(2) xor reg_t4(3) xor
           reg_t6(0) xor reg_t6(1) xor reg_t6(2) xor reg_t6(3) xor reg_t6(4) xor reg_t6(5);
    msg_auth_valid <= '1' when tag = b0 else '0';

    gReference: if G_REFERENCE = True generate
        signal to_t3_0      : std_logic_vector(128          -1 downto 0);
        signal to_t4_0      : std_logic_vector(128          -1 downto 0);
        signal to_t6_0      : std_logic_vector(128          -1 downto 0);

        signal t3_rkey      : std_logic_vector(128          -1 downto 0);
        signal t4_rkey      : std_logic_vector(128          -1 downto 0);
        signal t6_rkey      : std_logic_vector(128          -1 downto 0);
        signal t3_sel       : std_logic_vector(128          -1 downto 0);
        signal t4_sel       : std_logic_vector(128          -1 downto 0);
        signal t6_sel       : std_logic_vector(128          -1 downto 0);
        --! Input
        signal mat_Z0       : t_AES_state;
        signal mat_reg_t3_2 : t_AES_state;
        signal mat_reg_t4_3 : t_AES_state;
        signal mat_reg_t6_5 : t_AES_state;
        signal mat_reg_t3_0 : t_AES_state;
        signal mat_reg_t4_0 : t_AES_state;
        signal mat_reg_t6_0 : t_AES_state;
        signal mat_t3_rkey  : t_AES_state;
        signal mat_t4_rkey  : t_AES_state;
        signal mat_t6_rkey  : t_AES_state;
        --! Output
        signal mat_t3_0     : t_AES_state;
        signal mat_t3_1     : t_AES_state;
        signal mat_t4_0     : t_AES_state;
        signal mat_t4_1     : t_AES_state;
        signal mat_t6_0     : t_AES_state;
        signal mat_t6_1     : t_AES_state;

        signal c0, c1       : std_logic_vector(128          -1 downto 0);
        signal m0, m1       : std_logic_vector(128          -1 downto 0);
        signal msg0, msg1   : std_logic_vector(128          -1 downto 0);
        signal sel_output   : std_logic_vector(2            -1 downto 0);
    begin
        with sel_data select
        t3_sel <= b0                                     when "000",
                  Z0                                     when "010",
                  Z1                                     when "011",
                  ZEROS(63 downto 0) & len_a             when "111",
                  (others => '0')                        when others;

        with sel_data select
        t4_sel <= b1                                     when "000",
                  Z1                                     when "010",
                  Z0                                     when "011",
                  ZEROS(63 downto 0) & len_d             when "111",
                  (others => '0')                        when others;

        with sel_data select
        t6_sel <= b01                                    when "000",
                  Z0                                     when "010",
                  Z1                                     when "011",
                  ZEROS(63 downto 0) & (len_a xor len_d) when "111",
                  (others => '0')                        when others;

        t3_rkey <= reg_t3(0) xor t3_sel;
        t4_rkey <= reg_t4(0) xor t4_sel;
        t6_rkey <= reg_t6(0) xor t6_sel;

        u_Z0_map: entity work.AES_map(structure) port map (ii => Z0, oo => mat_Z0);

        --! T3
        --!     State -> Matrix
        u_t3_map1: entity work.AES_map(structure) port map (ii => reg_t3(0), oo => mat_reg_t3_0);
        u_t3_map2: entity work.AES_map(structure) port map (ii => reg_t3(2), oo => mat_reg_t3_2);
        u_t3_map3: entity work.AES_map(structure) port map (ii => t3_rkey,   oo => mat_t3_rkey );
        --!     Core
        u_t3_AES1:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t3_2, rkey => mat_t3_rkey, dout_fdb => mat_t3_0, dout => open);
        u_t3_AES2:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t3_0, rkey => mat_Z0, dout_fdb => mat_t3_1, dout => open);
        --!     Matrix -> State and Output
        u_t3_inv1: entity work.AES_invmap(structure) port map (ii => mat_t3_0, oo => to_t3_0);
        u_t3_inv2: entity work.AES_invmap(structure) port map (ii => mat_t3_1, oo => t3(1));
        t3(2) <= reg_t3(1);

        --! t4
        --!     State -> Matrix
        u_t4_map1: entity work.AES_map(structure) port map (ii => reg_t4(0), oo => mat_reg_t4_0);
        u_t4_map2: entity work.AES_map(structure) port map (ii => reg_t4(3), oo => mat_reg_t4_3);
        u_t4_map3: entity work.AES_map(structure) port map (ii => t4_rkey,   oo => mat_t4_rkey );
        --!     Core
        u_t4_AES1:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t4_3, rkey => mat_t4_rkey, dout_fdb => mat_t4_0, dout => open);
        u_t4_AES2:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t4_0, rkey => mat_Z0, dout_fdb => mat_t4_1, dout => open);
        --!     Matrix -> State and Output
        u_t4_inv1: entity work.AES_invmap(structure) port map (ii => mat_t4_0, oo => to_t4_0);
        u_t4_inv2: entity work.AES_invmap(structure) port map (ii => mat_t4_1, oo => t4(1));
        t4(2) <= reg_t4(1);
        t4(3) <= reg_t4(2);

        --! t6
        --!     State -> Matrix
        u_t6_map1: entity work.AES_map(structure) port map (ii => reg_t6(0), oo => mat_reg_t6_0);
        u_t6_map2: entity work.AES_map(structure) port map (ii => reg_t6(5), oo => mat_reg_t6_5);
        u_t6_map3: entity work.AES_map(structure) port map (ii => t6_rkey,   oo => mat_t6_rkey );
        --!     Core
        u_t6_AES1:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t6_5, rkey => mat_t6_rkey, dout_fdb => mat_t6_0, dout => open);
        u_t6_AES2:
        entity work.AES_Round(basic)
        generic map (G_SBOX_LOGIC => False)
        port map (din => mat_reg_t6_0, rkey => mat_Z0, dout_fdb => mat_t6_1, dout => open);
        --!     Matrix -> State and Output
        u_t6_inv1: entity work.AES_invmap(structure) port map (ii => mat_t6_0, oo => to_t6_0);
        u_t6_inv2: entity work.AES_invmap(structure) port map (ii => mat_t6_1, oo => t6(1));
        t6(2) <= reg_t6(1);
        t6(3) <= reg_t6(2);
        t6(4) <= reg_t6(3);
        t6(5) <= reg_t6(4);

        --! Output and new round
        c0 <= to_t3_0 xor t3(2) xor t4(1) xor (t6(3) and t4(3));
        c1 <= to_t6_0 xor t4(2) xor t3(1) xor (t6(5) and t3(2));
        m0 <= b0 xor c0;
        m1 <= b1 xor c1 xor m0;
        
        gLoop:
        for i in 128/8-1 downto 0 generate
            msg0(i*8+7 downto i*8) <= m0(i*8+7 downto i*8) when bdi_valid_bytes(i+128/8) = '1' else (others => '0');
            msg1(i*8+7 downto i*8) <= m1(i*8+7 downto i*8) when bdi_valid_bytes(i      ) = '1' else (others => '0');
        end generate;

        t3(0) <= to_t3_0 when sel_decout = '0' else to_t3_0 xor msg0;
        t4(0) <= to_t4_0 when sel_decout = '0' else to_t4_0 xor msg1;
        t6(0) <= to_t6_0 when sel_decout = '0' else to_t6_0 xor msg0 xor msg1;

        sel_output <= sel_tag & sel_decout;
        with sel_output select
        bdo(255 downto 128) <=  c0  when "00",
                                m0  when "01",
                                tag when others;
        bdo(127 downto   0) <= m1  when sel_decout = '1' else c1;
    end generate;
    gOptimized: if G_REFERENCE = False generate

    begin
        --! ========= Possible optimization via logic expansion
        --! b01     = b0 xor b1;
        --! pre_m0  = t3(2) xor (t6(3) and t4(3));
        --! pre_m1  = t4(2) xor (t6(5) and t3(2));
        --!
        --! -- Decryption
        --! m0      = b0 xor t3(0) xor t4(1) xor t3(2) xor (t6(3) and t4(3))
        --!         = b0 xor t3(0) xor t4(1) xor pre_m0
        --!         = b0 xor AES(reg_t3(2)) xor reg_t3(0)
        --!             xor AES(reg_t4(0)) xor Z0
        --!             xor pre_m0
        --! m1      = b0 xor t3(0) xor t4(1) xor t3(2) xor (t6(3) and t4(3)) xor
        --!         b1 xor t6(0) xor t3(1) xor t4(2) xor (t6(5) and t3(2))
        --!         = b01 xor t3(0) xor t4(1) xor pre_m1
        --!             xor t6(0) xor t3(1) xor pre_m0
        --!         = b01 xor AES(reg_t3(2)) xor reg_t3(0) xor AES(reg_t4(0)) xor Z0 xor pre_m1
        --!               xor AES(reg_t6(5)) xor reg_t6(0) xor AES(reg_t3(0)) xor Z0 xor pre_m0
        --!         = b01 xor AES(reg_t3(2)) xor reg_t3(0) xor AES(reg_t4(0))        xor pre_m1
        --!               xor AES(reg_t6(5)) xor reg_t6(0) xor AES(reg_t3(0))        xor pre_m0
        --!
        --! t3(0)   = b0           xor t4(1) xor t3(2) xor (t6(3) and t4(3))
        --!         = b0           xor t4(1) xor pre_m0
        --!         = b0           xor AES(reg_t4(0)) xor Z0 xor pre_m0
        --! t4(0)   = b0 xor t3(0) xor t4(1) xor t3(2) xor (t6(3) and t4(3)) xor
        --!           b1 xor t6(0) xor t3(1) xor t4(2) xor (t6(5) and t3(2))
        --!              xor t4(0)
        --!         = b01 xor t3(0) xor t4(1) xor pre_m1
        --!               xor t6(0) xor t3(1) xor pre_m0
        --!               xor t4(0)
        --!         = b01 xor AES(reg_t3(2)) xor reg_t3(0) xor AES(reg_t4(0))        xor pre_m1
        --!               xor AES(reg_t6(5)) xor reg_t6(0) xor AES(reg_t3(0))        xor pre_m0
        --!               xor AES(reg_t4(3)) xor reg_t4(0)
        --!
        --! t6(0)   = b0 xor t3(0) xor t4(1) xor t3(2) xor (t6(3) and t4(3)) xor
        --!                  t6(0)
        --!         = b0 xor t3(0) xor t4(1) xor pre_m0
        --!                  t6(0)
        --!         = b0 xor AES(reg_t3(2)) xor reg_t3(0) xor AES(reg_t4(0)) xor Z0 xor pre_m0
        --!                  AES(reg_t6(5)) xor reg_t6(0)
        --! -- Encryption
        --! c0      = t3(0) xor t3(2) xor t4(1) xor (t6(3) and t4(3))
        --!         = AES(reg_t3(2)) xor reg_t3(0) xor b0 xor AES(reg_t4(0)) xor Z0 xor pre_m0
        --!         = AES(reg_t3(2)) xor AES(reg_t4(0))
        --!            xor reg_t3(0) xor b0 xor Z0 xor pre_m0
        --! c1      = t6(0) xor t4(2) xor t3(1) xor (t6(5) and t3(2))
        --!         = AES(reg_t6(5)) xor reg_t6(0) xor b01 xor AES(reg_t3(0)) xor Z0 xor pre_m1
        --!         = AES(reg_t6(5)) xor AES(reg_t3(0))
        --!            xor reg_t6(0) xor b01 xor Z0 xor pre_m1
        --! t3(0)   = AES(reg_t3(2)) xor reg_t3(0) xor b0
        --! t4(0)   = AES(reg_t4(3)) xor reg_t4(0) xor b1
        --! t6(0)   = AES(reg_t6(5)) xor reg_t6(0) xor b01
        --! -- Encryption and Decryption
        --! t3(1)   = AES(reg_t3(0)) xor Z0
        --! t4(1)   = AES(reg_t4(0)) xor Z0
        --! t6(1)   = AES(reg_t6(0)) xor Z0

    end generate;


end architecture structure;